/********************************************************************
KWin - the KDE window manager
This file is part of the KDE project.

Copyright (C) 2014 Martin Gräßlin <mgraesslin@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/
#include <kwineffects.h>

namespace KWin
{

class FakeEffect : public Effect
{
    Q_OBJECT
public:
    FakeEffect() {}
    virtual ~FakeEffect() {}

    static bool supported() {
        return effects->isOpenGLCompositing();
    }

    static bool enabledByDefault() {
        return effects->property("testEnabledByDefault").toBool();
    }
};

} // namespace

KWIN_EFFECT_FACTORY_SUPPORTED_ENABLED( FakeEffectPluginFactory,
                                       KWin::FakeEffect,
                                       "fakeeffectplugin.json",
                                       return KWin::FakeEffect::supported();,
                                       return KWin::FakeEffect::enabledByDefault();)

#include "fakeeffectplugin.moc"
