/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2016 Martin Gräßlin <mgraesslin@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/
#ifndef WAYLANDCLIPBOARD_H
#define WAYLANDCLIPBOARD_H

#include <QObject>

class QThread;

namespace KWayland
{
namespace Client
{
class ConnectionThread;
class Seat;
class DataDeviceManager;
class DataDevice;
class DataSource;
}
}

class WaylandClipboard : public QObject
{
    Q_OBJECT
public:
    explicit WaylandClipboard(QObject *parent);
    ~WaylandClipboard();

private:
    void setup();
    QThread *m_thread;
    KWayland::Client::ConnectionThread *m_connectionThread;
    KWayland::Client::Seat *m_seat = nullptr;
    KWayland::Client::DataDeviceManager *m_dataDeviceManager = nullptr;
    KWayland::Client::DataDevice *m_dataDevice = nullptr;
    KWayland::Client::DataSource *m_dataSource = nullptr;
};

#endif
